#include <inttypes.h>
#include <avr/io.h>
#include <string.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include <avr/eeprom.h>


/*************************
 *     4GHZ GENER        *
 *  (C) OK1DX 2021       *
 *   ok1dx@amsat.org     *
 *************************
 
Free for amateur radio community
Modify as you like, just don'r remove original copyright.
No commerical usage allowed!!!

 version 1.0
 - first release
 version 2.0
 - added CW beacon, ref freq 10/25 MHz
 version 2.1
 - fix ADC ranges for keyboard
 */

// Xtal 11.059 MHz
#define F_CPU 11059400UL

// ADF clock [in kHz] - 25 MHz is the default, 10 MHz needs a XO change
// #define REFFREQ 10000
#define REFFREQ 25000

// FUSES:
// high byte 11001101 = RSTDISBL=1, WDTON=1, SPIEN=0, CKOPT=0, EESAVE=1, BOOTSZ1=1, BOOTSZ0=0, BOOTRST=1
// low byte  11101111 = BODLEVEL=1, BODEN=1, SUT1=1, SUT0=0 (4.1 msec), CKSEL0-3 = 1 (high freq xtal)


// LCD interface:
// PD0 - in, out D0
// PD1 - in, out D1
// PD2 - in, out D2
// PD3 - in, out D3
// PD4 - in, out D4
// PD5 - in, out D5
// PD6 - in, out D6
// PD7 - in, out D7
#define RS PC3  // 1 = data, 0 = instr
#define RW PC4  // 1 = read  0 = write
#define ENABLE PC5 // default 1

// ADF4351 interface:
#define ADF_LE PB0
#define ADF_DATA PB1
#define ADF_CLK PB2

// KEYBOARD interface:
// ADC7 - voltage sense (pullup 10kohm)

// Programming interface:
// PC6 - RESET
// PB3 - MOSI
// PB4 - MISO
// PB5 - SCK

// DDR bits 1=out 0=in
// PORT bits in input mode 1=pull up, 0=high Z

// PORT B - PB0-2 out ADF, PB3-5 in pullup (programmer), PB6,PB7 high Z XTAL
#define DDRB_ 0x07
#define PORTB_ 0x3F

// PORT C - PC0, PC2 n.c. in pullup, PC1 LED out, PC3-5 out LCD (all default 0), PC6 in pullup RESET
#define DDRC_ 0x3A
#define PORTC_ 0x45

// PORT D - output, LCD
#define DDRD_ 0xFF
#define PORTD_ 0x00

// initial ADF4351 register values (see datasheet)
#if (REFFREQ == 25000)
EEMEM uint32_t eeregs[6] = {0x002c0000, 0x0800e1a9, 0x00004e42, 0x000004b3, 0x008c803c, 0x00580005};
#elif (REFFREQ == 10000)
EEMEM uint32_t eeregs[6] = {0x006E0000, 0x0800a711, 0x00004642, 0x000004b3, 0x008c803c, 0x00580005};
#else
#error Unsupported reference freq
#endif
EEMEM uint32_t eefreq = 1000000;  // default eeprom freq 1 GHz
EEMEM uint8_t eepwr = 4;  // default level 5dBm
uint32_t regs[6];

// LCD display data
uint8_t line1[16];
uint8_t line2[16];

uint32_t khz = 0;  // frequency in kHz
uint8_t step = 0;  // index of freq step
#define MAXSTEP 6
uint8_t khzStep[7] = { 1, 2, 5, 10, 20, 50, 100 } ; // freq step table
uint8_t freqOK = 0; // flag freq sent to ADF

uint8_t pwr;  // output power level index (0=OFF, 1=-4dBm, 2=-1dBm, 3=2dBm, 4=5dBm, 5=beacon)
uint8_t div = 0;  // output divider index (0=div1, 1=div2, 2=div4, 3=div8,...)
uint8_t select = 0; // input data selector
#define SELFREQ 0
#define SELSTEP 1
#define SELPWR 2

uint16_t INT; // ADF INT value
uint16_t FRAC; // ADF FRACT value

#define CHAR_DEC 10
#define CHAR_INC 11
#define CHAR_OK 12
#define CHAR_C 13
#define CHAR_SEL 14
#define CHAR_NONE 15
#define CHAR_ERR 16

/* MORSE code table */
#define x0 0b111111
#define x1 0b111110
#define x2 0b111100
#define x3 0b111000
#define x4 0b110000
#define x5 0b100000
#define x6 0b100001
#define x7 0b100011
#define x8 0b100111
#define x9 0b101111

#define xA 0b110
#define xB 0b10001
#define xC 0b10101
#define xD 0b1001
#define xE 0b10
#define xF 0b10100
#define xG 0b1011
#define xH 0b10000
#define xI 0b100
#define xJ 0b11110
#define xK 0b1101
#define xL 0b10010
#define xM 0b111
#define xN 0b101
#define xO 0b1111
#define xP 0b10110
#define xQ 0b11011
#define xR 0b1010
#define xS 0b1000
#define xT 0b11
#define xU 0b1100
#define xV 0b11000
#define xW 0b1110
#define xX 0b11001
#define xY 0b11101
#define xZ 0b10011

#define plus 0b101010
#define slash 0b101001
#define eqsign 0b110001
#define minus 0b1100001
#define quest 0b1001100
#define comma 0b1110011
#define dot 0b1101010
#define zavorka1 0b101101
#define zavorka2 0b1101101
#define space 0xFE
#define end 0xFF

uint8_t cwreg;
uint8_t * eeptr; // pointer to message
uint8_t cwkey; // 0=key up, 1=key down
uint8_t excwkey; // last key value (change detect)
volatile uint8_t timer;
uint8_t tim1;  // local copy of timer
EEMEM uint8_t eemsg[] = { xO, xK, x1, xD, xX, space, xB, xE, xA, xC, xO, xN, space, end };  // message to be sent
#define CWSPD 60; // CW speed factor. length of dot in msec. Max 85

void lcdEnPulse()
{
	PORTC |= (1<<ENABLE); // enable pulse
	_delay_loop_1(5);  // small delay
	PORTC &= ~(1<<ENABLE);
	_delay_loop_1(5);  // small delay
}

// writes previously set data to LCD registers (selected previously)
// write pulse, then wait till BF == 0 (read status)
void lcdOut()
{
    uint8_t status;
	PORTC &= ~(1<<RW); // write mode

	lcdEnPulse();

    DDRD = 0;  // all pins inputs
	PORTC |= (1<<RW);  // read mode
	PORTC &= ~(1<< RS);  // command
	_delay_loop_1(5);
	do
	{
        PORTC |= (1<<ENABLE);
        _delay_loop_1(5);
	    status = PIND;
		PORTC &= ~(1<<ENABLE);
		_delay_loop_1(5);
	}
	while ((status & 0x80) != 0);

	PORTC &= ~(1<<RW); // write mode
	DDRD = 0xFF; // outputs again
}

void lcdData(uint8_t data)
{
	PORTD = data;
	PORTC |= 1<< RS;
	lcdOut();
}

void lcdCmd(uint8_t cmd)
{
	PORTD = cmd;
	PORTC &= ~(1<< RS);
	lcdOut();
}

void lcdLine1()
{
 	uint8_t arrow = 'v';

	line1[2] = step + '0';

     if (select == SELPWR)
	 	arrow = '>';
	 else if (select == SELSTEP)
	 	arrow = '<';
     line1[4] = arrow;

	switch(pwr)
	{
	case 1:
		memcpy(&(line1[10]),"-4 dBm",6);
		break;
	case 2:
		memcpy(&(line1[10]),"-1 dBm",6);
		break;
    case 3:
		memcpy(&(line1[10]),"2 dBm ",6);
		break;
	case 4:
		memcpy(&(line1[10]),"5 dBm ",6);
		break;
	case 5:
	    memcpy(&(line1[10]),"BEACON",6);
		break;
	default:
		memcpy(&(line1[10]),"OFF   ",6);
		break;
	}

    uint8_t* lineptr = line1;
	lcdCmd(0x80); //1st line, position 0
	uint8_t ctr = 16;  // 16 characters on line
	while(ctr > 0)
	{
		lcdData(*lineptr);
		lineptr++;
		ctr--;
	}
}

 void lcdLine2()
 {
	 uint32_t f1 = khz;
	 uint8_t ctr = 11;
	 uint8_t znak;

	 while (f1 > 0)
	 {
		 int n = f1 % 10;
		 f1 /= 10;
		 line2[ctr] = n + '0';
		 ctr--;
		 if (ctr == 8)
		 {
			 line2[ctr] = '.';
			 ctr--;
		 }
	 }
	 while (ctr >=3)
	 {
		 if (ctr == 8)
		 {
			  znak = '.';
		 }
		 else if (ctr >= 7)
		 { 
			 znak = '0';
		 }
		 else
		 {
			 znak = ' ';
		 }
		 line2[ctr] = znak;
		 ctr--;
	 }

	uint8_t* lineptr = line2;
	lcdCmd(0xC0); //2nd line, position 0
	ctr = 16;  // 16 characters on line
	while(ctr > 0)
	{
		lcdData(*lineptr);
		lineptr++;
		ctr--;
	}
 }

// update data regs contest for given freq and output level
 void UpdateRegs()
 {
    uint8_t pwr2;
	uint32_t temp32;

    // register 0 - INT and FRAC
    temp32 = INT;
	temp32 = temp32 << 12;
	temp32 += (FRAC & 0x0FFF);
	temp32 = temp32 << 3;
	regs[0] = temp32;

	// register 4 - DIV, PWR and Out enable
	div &= 7; // 3 bit length
	regs[4] &= 0x008FFFC7;  // reset bits
	temp32= div;  // output divider
	temp32 = temp32 << 20;
	if (pwr != 0)
	{
	   pwr2 = (pwr-1) & 0x03; // 2 bits mask
	   if (pwr==5)
	   {
	     pwr2 = 3;  // beacon with max power
	   }
	   temp32 |= (1<<5) | (pwr2<<3);  // out enabled, level
	}
	regs[4] |= temp32;
 }

// send 4 bytes from data as register tlg to ADF, Adr points to LSB
 void WriteReg(uint32_t* adr32)
 {
 	uint8_t ctr = 32;
	uint8_t reg = 0;
	uint8_t* adr;

	adr = (uint8_t*)adr32 + 3; // point to MSB byte
	
    PORTB &= ~(1<<ADF_LE);  // LE = 0
	PORTB &= ~(1<<ADF_CLK);  // clk = 0

	while (ctr > 0)
	{
		if ((ctr & 0x07) == 0) // every 8th cycle
		{
			reg = *adr;
			adr--;
		}

		if (reg&0x80) // take MSB bit
			PORTB |= (1<<ADF_DATA);  // data = 1
        else
		    PORTB &= ~(1<<ADF_DATA); // data = 0
        reg += reg;  // bit shift
		PORTB |= (1<<ADF_CLK);  // rising edge = strobe now
		ctr--;
		PORTB &= ~(1<<ADF_CLK);  // clk = 0
    }
	PORTB |= 1<<ADF_LE;     // LE=1
 }

// write freq and pwr into EEPROM
 void WriteEEPROMdata()
 {
   eeprom_write_dword (&eefreq, khz);
   eeprom_write_byte(&eepwr, pwr);
 }

 void cwtx(void)
 {
	cli();
	tim1 = timer;  // local copy of timer
	sei();

	if (tim1 == 0)		// timer expired
	{
		if (cwreg > 0)		// still something to send
		{
			if (cwreg == space)	// sending "space" symbol
			{
				cwkey = 0;	// key up
				tim1 = 2 * CWSPD;
				cwreg = 0;		// symbol sent
			}
			else
			{
				if (cwkey)
				{	
					cwkey = 0;	// key up
					tim1 = CWSPD;
					if (cwreg == 1)
					{
						tim1 = 3* CWSPD;
						cwreg = 0;	// symbol sent, sending space between symbols
					}
				}
				else
				{
					cwkey = 1;	// key down
					if (cwreg & 1)
					{
						tim1 = 3*CWSPD;	// dash
					}
					else
					{
						tim1 = CWSPD;		// dot
					}
					cwreg = cwreg >> 1;
				}
			}
		}

		cli();
		timer = tim1;  // return the volatile timer 
		sei();
	
		if (cwreg == 0)
		{
		    // take next symbol
			cwreg = eeprom_read_byte(eeptr);
			eeptr++;
			if (cwreg == end)
			{
			   eeptr = eemsg;  // again from msg begin
			   cwreg = eeprom_read_byte(eeptr);
			   eeptr++;
			}
		}

		if (cwkey != excwkey)
		{
		   regs[4] &= 0x00FFFFDF;  // reset rf output enable bit - output off
		   if (cwkey)
		   { 
		     PORTC |= (1<<PC1);  // CW LED on
			 if (pwr == 5)  // by beacon only
			 {
			   regs[4] |= (1<<5);  // output on
			 }
		   }
		   else
		   {
		     PORTC &= ~(1<<PC1);  // CW LED off
		   }

		   if (pwr == 5)  // by beacon only
		   {
		      WriteReg(&regs[4]); // update ADF register 4
		   }

		   excwkey = cwkey;
		}
	}
}

// convert scan code (ADC value) into key number
// limits given by resistor net values (divider, 10kohm to Vcc)
 uint8_t keybDecode(uint16_t scancode)
 {
    uint8_t scanchar;

	if ((scancode >= 70) && (scancode <= 116))  // R=1kohm
	   scanchar = 9;
    else if ((scancode >= 151) && (scancode <= 190)) // R=2kohm
	   scanchar = 8;
    else if ((scancode >= 220) && (scancode <= 252)) // R=3kohm
	   scanchar = 7;
    else if ((scancode >= 278) && (scancode <= 306)) // R=4kohm
	   scanchar = CHAR_INC;
    else if ((scancode >= 329) && (scancode <= 353)) // R=5kohm
	   scanchar = CHAR_C;
    else if ((scancode >= 373) && (scancode <= 394)) // R=6kohm
	   scanchar = 6;
    else if ((scancode >= 412) && (scancode <= 431)) // R=7kohm
	   scanchar = 5;
    else if ((scancode >= 469) && (scancode <= 500)) // R=9kohm
	   scanchar = 4;
    else if ((scancode >= 523) && (scancode <= 549)) // R=11kohm
	   scanchar = CHAR_DEC;
    else if ((scancode >= 568) && (scancode <= 589)) // R=13kohm
	   scanchar = CHAR_OK;
    else if ((scancode >= 605) && (scancode <= 623)) // R=15kohm
	   scanchar = 3;
    else if ((scancode >= 637) && (scancode <= 652)) // R=17kohm
	   scanchar = 2;
    else if ((scancode >= 664) && (scancode <= 677)) // R=19kohm
	   scanchar = 1;
    else if ((scancode >= 697) && (scancode <= 715)) // R=22.3kohm
	   scanchar = 0;
    else if ((scancode >= 728) && (scancode <= 743)) // R=25.6kohm
	   scanchar = CHAR_SEL;
    else if (scancode >= 1000) // R=infinite, no button pressed
	   scanchar = CHAR_NONE;
    else
	   scanchar = CHAR_ERR;

	return scanchar;
 }

// loop waiting for valid keyboard character.
 uint8_t getKeyboard()
 {
  uint8_t ctr = 0;
  uint8_t keyflag = 0; // set when key released
  uint8_t scanchar;
  uint8_t retchar = CHAR_NONE;
  uint16_t scancode;

  do
  {
  	ADCSRA |= (1<<ADSC);  // command start convertion
	do
	{
	   cwtx(); // transmit cw
	}
	while (ADCSRA & (1<<ADSC));  // wait till end of convertion
    ADCSRA |= (1<<ADIF);

	scancode = ADCL;  // read ADC, LSB must be first
	scancode |= (ADCH<<8);

	scanchar = keybDecode(scancode);

    if (scanchar == CHAR_NONE)
	{
	   // key must preceeded no key pressed state
	   keyflag = 0xFF;
	   ctr = 0;
	   continue;
	}

	if ((scanchar != CHAR_ERR) && keyflag)
	{
		if (scanchar == retchar)
		{
		   if (ctr < 0xFF)
			 ctr++;
		}
		else
		{
			retchar = scanchar;
			ctr = 0;
		}
	}
  }
  // the key must be the same for at least certain cycles
  while ((ctr < 4) || !keyflag);
  return retchar;
 }

// add one digit to frequency (from keyboard)
 void khzIn(char x)
 { 
     if (x < 10)
	 {
	     if (freqOK)
		 {
		    khz = 0;  // clear freq upon first entry
			freqOK = 0;
		 }

		 khz = 10 * khz + x;
		 if (khz > 4500000)
			 khz = x;
	 }
 }

//  calculate ADF frequency setting parameters
 void CalcADF()
 {
	 div = 0;
	 uint32_t vco = khz;
	 while (vco < 2200000)
	 {
		 vco *= 2;
		 div++;
	 }

     INT = 0;
	 while (vco >= REFFREQ)
	 {
		 vco -= REFFREQ;
		 INT++;
	 }
	 FRAC = vco >> 3;

	 khz = (REFFREQ * (uint32_t)INT + (FRAC << 3)) >> div;
 }

// button Increment
 void butInc()
 {
	 uint32_t x;
	 switch (select)
	 {
	 case SELFREQ:
		 x = khzStep[step];
		 if (div < 3)
		   x = x << (3-div);
		 khz += x;
		 break;
	 case SELSTEP:
		 if (step < MAXSTEP)
			 step++;
		 else
			 step = 0;
		 break;
	 case SELPWR:
		 if (pwr < 5)  // horni zarazka
			 pwr++;
		 break;
	 }
 }

// button Decrement
 void butDec()
 {
	 uint32_t x;
	 switch (select)
	 {
	 case SELFREQ:
		 x = khzStep[step];
		 if (div < 3)
			x = x << (3-div);
		 khz -= x;
		 break;
	 case SELSTEP:
		 if (step > 0)
			 step--;
		 else
			 step = MAXSTEP;
		 break;
	 case SELPWR:
		 if (pwr > 0)  // dolni zarazka
			 pwr--;
		 break;
	 }
 }

// button Select entry mode
 void butSelect()
 {
	 if (select < 2)
		select++;
	 else
		select = 0;
 }

// button OK (frequency confirmation - send it to ADF)
 void butOK()
 {
 	uint8_t i;
	 if (khz < 35000)
	 {
		 // freq moc nizka
		 khz = 0;
		 return;
	 }

	 CalcADF();
	 UpdateRegs();
	 // send all regs to ADF (order R5, R4, R3, R2, R1, R0). ToDo: send only needed regs (R1, R2, R3 and R5 are constant)
     for(i=0; i<6; i++)
	   WriteReg(&regs[5-i]);
     freqOK = 0xFF;
 }

void init()
{
    // initiate IO ports
	DDRB = DDRB_;
	PORTB = PORTB_;
	DDRC = DDRC_;
	PORTC = PORTC_;
	DDRD = DDRD_;
	PORTD = PORTD_;

    // ADC init - clock/128 (conversion 15 cycles -> 0.2 msec), ADC enabled
	ADCSRA = (1<<ADPS0) | (1<<ADPS1) | (1<<ADPS2) | (1<<ADEN);
    // using ADC7 with reference from Vcc
	ADMUX = 7 | (1<<REFS0);

	// LCD init in 8bit mode
	// clock = 0.1 usec, value = 4*clock
	// RS, RW, ENABLE = 0 (init value)
	_delay_loop_2(50000);  // 20 msec
	_delay_loop_2(50000);  // 20 msec
	PORTD = 0x30;
	lcdEnPulse();
	_delay_loop_2(12500); // 5 msec
	lcdEnPulse();
	_delay_loop_2(400);  // 160 usec
	lcdEnPulse();
	_delay_loop_2(100);  // 40 usec
	lcdCmd(0x38);  // // 8 bit, 2 lines, 5x7 mode
	lcdCmd(0x08);  // display off
	lcdCmd(0x01);  // display clear
	lcdCmd(0x06);  // increment, cursor move
	lcdCmd(0x0C);  // display ON, no cursor, no blink
	// lcdCmd(0x01);  // display clear

    // copy initial ADF register values from EEPROM to data memory
    eeprom_read_block(regs, eeregs, 24);

	// recall frequency and power level
	khz = eeprom_read_dword (&eefreq);
    pwr = eeprom_read_byte(&eepwr);
	// khz = 1000000;
	// pwr = 4;

	// LCD display RAM buffer
	memcpy(line1,"S=0    P= -2 dBm",16);
	memcpy(line2,"F=           MHz",16);

    // cw keyer
    eeptr = eemsg;  // start of CW messagetext
	cwreg = 0;
	cwkey = 0;
	excwkey = 0;

	TCCR0 = (1<< CS00) | (1<<CS01);	// prescaler 64
    TIMSK = 1<<TOIE0;		// enable timer 0 overflow IRQ
    sei();
}


int main (void)
{
	init();

	butOK();  // adjust all for init frequency

	while(1)
	{
	
		lcdLine1();  // ToDo refresh only needed line
		lcdLine2();

	    uint8_t key = getKeyboard();
		switch (key)
		{
			case CHAR_OK:
			    WriteEEPROMdata();
				butOK();
				break;
			case CHAR_C:
				khz = 0;
				break;
			case CHAR_INC:
				butInc();
				if (select != SELSTEP)
					butOK();
				break;
			case CHAR_DEC:
				butDec();
				if (select != SELSTEP)
					butOK();
				break;
			case CHAR_SEL:
				butSelect();
				break;
			default:
				khzIn(key);
				break;
		}
	}
}

// timer decremented every 1 msec
// clock = 11 MHz, prescaller 64, preset value 172 =  -> INT every 1 msec
ISR(TIMER0_OVF_vect)
{
	TCNT1 = 84;  // 256-172, counter incremented
	if (timer > 0)
		timer--;
}
